# coding=utf-8

"""PKM file structure for Generation 5.

All PKM Struct modules should declare a `pkm_struct` variable that is
an instance of construct.Struct.

Note about BitStructs: it appears Construct has some odd formatting, so
the bit order is reversed when declaring BitStructs. This should be
noticeable for ribbons.

@see http://projectpokemon.org/wiki/Pokemon_Black/White_NDS_Structure
"""

__author__ = 'Patrick Jacobs <ceolwulf@gmail.com>'

# I hate doing this, but apparently it's Construct convention
from construct import *
from pypkm.util import Swapped
from pypkm.adapters.gen5 import NicknameAdapter, OTNameAdapter

_block0 = Struct('_block0',
    ULInt32('pv'),
    Padding(2),
    ULInt16('checksum'),
)

_blockA = Struct('_blockA',
    ULInt16('id'),
    ULInt16('item'),
    ULInt16('ot_id'),
    ULInt16('ot_secret_id'),
    ULInt32('exp'),
    ULInt8('happiness'),
    ULInt8('ability'),
    BitStruct('markings',
        Padding(2),
        Flag('diamond'),
        Flag('star'),
        Flag('heart'),
        Flag('square'),
        Flag('triangle'),
        Flag('circle'),
    ),
    ULInt8('language'),
    Struct('evs',
        ULInt8('hp'),
        ULInt8('attack'),
        ULInt8('defense'),
        ULInt8('speed'),
        ULInt8('spattack'),
        ULInt8('spdefense'),
    ),
    Struct('cvs',
        ULInt8('cool'),
        ULInt8('beauty'),
        ULInt8('cute'),
        ULInt8('smart'),
        ULInt8('tough'),
        ULInt8('sheen'),
    ),
    Swapped(BitStruct('sinnoh_ribbons_set1',
        Flag('royal'),
        Flag('gorgeous'),
        Flag('smile'),
        Flag('snooze'),
        Flag('relax'),
        Flag('careless'),
        Flag('downcast'),
        Flag('shock'),
        Flag('alert'),
        Flag('world_ability'),
        Flag('pair_ability'),
        Flag('multi_ability'),
        Flag('double_ability'),
        Flag('great_ability'),
        Flag('ability'),
        Flag('sinnoh_champ'),
    )),
    Swapped(BitStruct('sinnoh_ribbons_set2',
        Padding(4),
        Flag('premiere'),
        Flag('classic'),
        Flag('carnival'),
        Flag('festival'),
        Flag('blue'),
        Flag('green'),
        Flag('red'),
        Flag('legend'),
        Flag('history'),
        Flag('record'),
        Flag('footprint'),
        Flag('gorgeous_royal'),
    )),
)

_blockB = Struct('_blockB',
    Struct('moves',
        ULInt16('move1'),
        ULInt16('move2'),
        ULInt16('move3'),
        ULInt16('move4'),
        ULInt8('move1_pp'),
        ULInt8('move2_pp'),
        ULInt8('move3_pp'),
        ULInt8('move4_pp'),
        ULInt8('move1_ppups'),
        ULInt8('move2_ppups'),
        ULInt8('move3_ppups'),
        ULInt8('move4_ppups'),
    ),
    Embed(Swapped(BitStruct('x38',
        Flag('is_nicknamed'),
        Flag('is_egg'),
        Struct('ivs',
            BitField('spdefense', 5),
            BitField('spattack', 5),
            BitField('speed', 5),
            BitField('defense', 5),
            BitField('attack', 5),
            BitField('hp', 5),
        ),
    ))),
    Swapped(BitStruct('hoenn_ribbons_set1',
        Flag('smart_master'),
        Flag('smart_hyper'),
        Flag('smart_super'),
        Flag('smart'),
        Flag('cute_master'),
        Flag('cute_hyper'),
        Flag('cute_super'),
        Flag('cute'),
        Flag('beauty_master'),
        Flag('beauty_hyper'),
        Flag('beauty_super'),
        Flag('beauty'),
        Flag('cool_master'),
        Flag('cool_hyper'),
        Flag('cool_super'),
        Flag('cool'),
    )),
    Swapped(BitStruct('hoenn_ribbons_set2',
        Flag('world'),
        Flag('earth'),
        Flag('national'),
        Flag('country'),
        Flag('sky'),
        Flag('land'),
        Flag('marine'),
        Flag('effort'),
        Flag('artist'),
        Flag('victory'),
        Flag('winning'),
        Flag('champion'),
        Flag('tough_master'),
        Flag('tough_hyper'),
        Flag('tough_super'),
        Flag('tough'),
    )),
    Embed(BitStruct('x40',
        BitField('alt_form_unshifted', 5), # leftshift 3
        Flag('is_genderless'), # if both False, then it's male
        Flag('is_female'),
        Flag('is_fateful'),
    )),
    ULInt8('nature'),
    Flag('has_dwability'),
    Padding(5),
)

_blockC = Struct('_blockC',
    NicknameAdapter(StrictRepeater(11, ULInt16('nickname'))),
    Padding(1),
    ULInt8('hometown'),
    Swapped(BitStruct('sinnoh_ribbons_set3',
        Flag('smart_master'),
        Flag('smart_ultra'),
        Flag('smart_great'),
        Flag('smart'),
        Flag('cute_master'),
        Flag('cute_ultra'),
        Flag('cute_great'),
        Flag('cute'),
        Flag('beauty_master'),
        Flag('beauty_ultra'),
        Flag('beauty_great'),
        Flag('beauty'),
        Flag('cool_master'),
        Flag('cool_ultra'),
        Flag('cool_great'),
        Flag('cool'),
    )),
    BitStruct('sinnoh_ribbons_set4',
        Padding(4),
        Flag('tough_master'),
        Flag('tough_ultra'),
        Flag('tough_great'),
        Flag('tough'),
    ),
    Padding(5),
)

_blockD = Struct('_blockD',
    OTNameAdapter(StrictRepeater(8, ULInt16('ot_name'))),
    Struct('egg_date',
        ULInt8('year'), # minus 2000
        ULInt8('month'),
        ULInt8('day'),
    ),
    Struct('met_date',
        ULInt8('year'), # minus 2000
        ULInt8('month'),
        ULInt8('day'),
    ),
    ULInt16('egg_location'),
    ULInt16('met_location'),
    ULInt8('pokerus'), # needs additional calculations
    ULInt8('ball'),
    Embed(BitStruct('x84',
        Flag('ot_is_female'),
        BitField('met_level', 7),
    )),
    ULInt8('encounter_type'),
    Padding(2),
)

# Battle data
_blockE = Struct('_blockE',
    BitStruct('status',
        Flag('toxic'),
        Flag('paralyzed'),
        Flag('frozen'),
        Flag('burned'),
        Flag('poisoned'),
        BitField('asleep_rounds', 3),
    ),
    Byte('x89'),
    Padding(2),
    ULInt8('level'),
    ULInt8('capsule_index'),
    Struct('stats',
        ULInt16('current_hp'),
        ULInt16('max_hp'),
        ULInt16('attack'),
        ULInt16('defense'),
        ULInt16('speed'),
        ULInt16('spattack'),
        ULInt16('spdefense'),
    ),
    Bytes('trash_data', 56),
    Padding(8)
)

pkm_struct = Struct('pkm_struct',
    Embed(_block0),
    Embed(_blockA),
    Embed(_blockB),
    Embed(_blockC),
    Embed(_blockD),
)

pkm_party_struct = Struct('pkm_party_struct',
    Embed(pkm_struct),
    Embed(_blockE),
)

# Data sent from the GTS server to the client
pkm_gtsserver_struct = Struct('pkm_gtsserver_struct',
    Bytes('encrypted_pkm', 220),
    Padding(16),
    ULInt16('id'),
    ULInt8('gender'),
    ULInt8('level'),
    Struct('requested',
        ULInt16('id'),
        ULInt8('gender'),
        ULInt8('min_level'),
        ULInt8('max_level'),
    ),
    Padding(1),
    ULInt8('ot_gender'),
    Padding(1),
    Struct('deposited_time',
        ULInt16('year'),
        ULInt8('month'),
        ULInt8('day'),
        ULInt8('hour'),
        ULInt8('minute'),
        ULInt8('second'),
        Padding(1),
    ),
    Struct('traded_time',
        ULInt16('year'),
        ULInt8('month'),
        ULInt8('day'),
        ULInt8('hour'),
        ULInt8('minute'),
        ULInt8('second'),
        Padding(1),
    ),
    ULInt32('pv'),
    ULInt16('ot_id'),
    ULInt16('ot_secret_id'),
    OTNameAdapter(StrictRepeater(8, ULInt16('ot_name'))),
    ULInt8('country'),
    ULInt8('city'),
    ULInt8('ot_sprite'),
    Flag('is_exchanged'),
    ULInt8('version'),
    ULInt8('language'),
    Padding(2),
)

# Data sent from the GTS client to the server
pkm_gtsclient_struct = Struct('pkm_gtsclient_struct',
    UBInt32('checksum'),
    ULInt32('pv'),
    ULInt16('length'),
    Padding(2),
    Bytes('encrypted_pkm', 220),
    Padding(16),
    ULInt16('id'),
    ULInt8('gender'),
    ULInt8('level'),
    Struct('requested',
        ULInt16('id'),
        ULInt8('gender'),
        ULInt8('min_level'),
        ULInt8('max_level'),
    ),
    Padding(1),
    ULInt8('ot_gender'),
    ULInt8('ot_nature'),
    Padding(4),
    Struct('deposited_time',
        ULInt16('year'),
        ULInt8('month'),
        ULInt8('day'),
        ULInt8('hour'),
        ULInt8('minute'),
        ULInt8('second'),
        Padding(1),
    ),
    Struct('traded_time',
        ULInt16('year'),
        ULInt8('month'),
        ULInt8('day'),
        ULInt8('hour'),
        ULInt8('minute'),
        ULInt8('second'),
        Padding(1),
    ),
    ULInt16('ot_id'),
    ULInt16('ot_secret_id'),
    OTNameAdapter(StrictRepeater(8, ULInt16('ot_name'))),
    ULInt8('country'),
    ULInt8('city'),
    ULInt8('ot_sprite'),
    Flag('is_exchanged'), # always False
    ULInt8('version'),
    ULInt8('language'),
    ULInt8('unknown_0to8'),
    ULInt8('tower_floors'),
    Padding(4),

    # according to maxg, the 128-byte signature is created by
    # sending the party struct to pkvldtprod.nintendo.co.jp and
    # using the token returned by the server (if it passes
    # validation). because it would be a total PITA to either
    # reverse engineer or include the legitimate validation, and
    # that this library is made with fake GTS servers in mind, I'm
    # going to default to NUL bytes. fake GTS servers can't even
    # validate the signature, so it shouldn't matter.
    Bytes('struct_signature', 128),
    ULInt32('terminator'), # always 128
)